"""Helper function for interaction with the underlying operating system."""

import glob
import os
import shutil
from typing import Iterator


def remove_file(file_path: str) -> None:
    """Remove file or directory including all its contents."""
    if os.path.isfile(file_path):
        os.remove(file_path)
    else:
        shutil.rmtree(file_path)


def multi_glob(*patterns: str) -> Iterator[str]:
    """Perform glob search on all arguments."""
    for pattern in patterns:
        for file_path in glob.glob(pattern):
            yield file_path



"""This module provides validation of Incoming requests."""

import functools
import json
import pathlib
from typing import Any, Callable, Dict, List

import fastjsonschema

import flask


schemas_path = pathlib.Path(__file__).parent / 'schemas'


def request_schema(schema_name: str) -> Callable:
    """Validate current request payload with given json schema.

    `request_schema` can be used only as a flask endpoint decorator. Must be
    called within request scope.
    """
    def request_schema_decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def request_schema_wrapper(*args: List[Any], **kwargs: Dict[str, Any]) -> Any:
            try:
                payload = flask.request.get_json()
                schema_path = get_schema_path(schema_name)
                validator = compile_json_schema(schema_path)
                validator(payload)
                return func(*args, **kwargs)
            except fastjsonschema.JsonSchemaException as e:
                flask.abort(400, e.message)
        return request_schema_wrapper
    return request_schema_decorator


@functools.lru_cache
def compile_json_schema(schema_path: pathlib.Path) -> Callable:
    """Compiles given schema to fastjson validation function."""
    with open(schema_path.absolute()) as schema_file:
        schema = json.load(schema_file)
        return fastjsonschema.compile(schema)


def get_schema_path(schema_name: str) -> pathlib.Path:
    """Return path to json schema file."""
    return schemas_path / schema_name



"""Trickster is a service providing REST API configurable by REST API."""


class TricksterException(Exception):
    """Exception from Trickster."""


class RouteConfigurationError(TricksterException, ValueError):
    """Raised when route could not be configured because of value error."""

    http_code: int = 400


class DuplicateRouteError(RouteConfigurationError):
    """Raised when route could not be configured because there is a duplicatit in ids."""

    http_code: int = 409


class MissingRouteError(RouteConfigurationError):
    """Raised when route could not be configured because it doesn't exist."""

    http_code: int = 404


class AuthenticationError(TricksterException):
    """Exception raised when user could not be authenticated."""



"""User-configurable routing."""

from __future__ import annotations

import enum
import json
import random
import re
import time
import uuid
from typing import Any, Dict, Iterable, List, Optional, Union

import flask

from trickster import DuplicateRouteError, MissingRouteError, RouteConfigurationError
from trickster.auth import Auth
from trickster.collections import IdItem, IdList
from trickster.input import IncomingRequest


class Delay:
    """Delay allows you to se arbitrary time to wait before response is returned."""

    def __init__(self, min_delay: float = 0.0, max_delay: float = 0.0):
        if min_delay > max_delay:
            raise RouteConfigurationError(
                f'Minimum delay cannot be longer than maximum delay: {min_delay} > {max_delay}'
            )
        self.min_delay = min_delay
        self.max_delay = max_delay

    def serialize(self) -> Union[Optional[List[float]], float]:
        """Convert Delay to json."""
        if self.min_delay == self.max_delay:
            return self.min_delay  # Express delay as one number
        return [self.min_delay, self.max_delay]

    def wait(self) -> None:
        """Put program to sleep for random amount of time withing the specified range."""
        wait_time = random.uniform(self.min_delay, self.max_delay)
        time.sleep(wait_time)

    @classmethod
    def deserialize(cls, data: Union[Optional[List[float]], float]) -> Delay:
        """Convert json to Delay."""
        if data is None:
            return cls(0.0, 0.0)
        if isinstance(data, (int, float)):
            return cls(data, data)
        return cls(*data)


class ResponseSelectionStrategy(enum.Enum):
    """Strategy of how to select a Response from list of responses."""

    cycle = 'cycle'
    random = 'random'
    greedy = 'greedy'

    def select_response_cycle(self, responses: List[Response]) -> Optional[Response]:
        """Select proper response from list of candidate responses.

        Consumes responses in order of definition. Cycles through items one by one.
        """
        candidate = None

        for response in responses:
            if response.is_active and (candidate is None or response.used_count < candidate.used_count):
                candidate = response
        return candidate

    def select_response_random(self, responses: List[Response]) -> Optional[Response]:
        """Select proper response from list of candidate responses.

        Selects random response from all available.
        """
        population = [r for r in responses if r.is_active]
        weights = [r.weight for r in population]
        result = random.choices(population=population, weights=weights, k=1)
        return result[0] if result else None

    def select_response_greedy(self, responses: List[Response]) -> Optional[Response]:
        """Select proper response from list of candidate responses.

        Consumes responses in order of definition until the first one is exhausted,
        then starts consuming the next in the row.
        """
        for response in responses:
            if response.is_active:
                return response
        return None

    def select_response(self, responses: Iterable[Response]) -> Optional[Response]:
        """Select proper response from list of candidate responses."""
        method_name = f'select_response_{self.value}'
        method = getattr(self, method_name)
        return method(responses)

    def serialize(self) -> str:
        """Convert ResponseSelectionStrategy to json."""
        return self.value

    @classmethod
    def deserialize(cls, method: Optional[str] = None) -> ResponseSelectionStrategy:
        """Convert json to ResponseSelectionStrategy."""
        return cls(method or 'greedy')


class Response(IdItem):
    """Container for predefined response."""

    def __init__(
        self,
        id: str,
        body: Any,
        delay: Delay,
        headers: Optional[Dict[str, Any]] = None,
        status: int = 200,
        repeat: Optional[int] = None,
        weight: float = 0.5
    ) -> None:
        super().__init__(id)
        self.body = body
        self.headers = headers or {}
        self.status = status
        self.repeat = repeat
        self.weight = weight
        self.delay = delay
        self.used_count = 0

    @property
    def serialized_body(self) -> str:
        """Convert specified response body to string."""
        if isinstance(self.body, str):
            return self.body
        else:
            return json.dumps(self.body)

    def as_flask_response(self) -> flask.Response:
        """Convert Request to flask.Response suitable to return from an endpoint."""
        return flask.Response(
            response=self.serialized_body,
            status=self.status,
            headers=self.headers
        )

    def serialize(self) -> Dict[str, Any]:
        """Convert Response to json."""
        return {
            **super().serialize(),
            'is_active': self.is_active,
            'status': self.status,
            'used_count': self.used_count,
            'weight': self.weight,
            'headers': self.headers,
            'repeat': self.repeat,
            'delay': self.delay.serialize(),
            'body': self.body
        }

    def use(self) -> None:
        """Increases usage counter of Response."""
        self.used_count += 1

    @property
    def is_active(self) -> bool:
        """Return True if response has some uses left."""
        return self.repeat is None or self.repeat > self.used_count

    def wait(self) -> None:
        """Sleep for time specified in the response."""
        self.delay.wait()

    @classmethod
    def deserialize(cls, data: Dict[str, Any]) -> Response:
        """Convert json to Response."""
        delay = Delay.deserialize(data.pop('delay', None))

        if 'headers' in data:
            headers = data.pop('headers')
        elif not isinstance(data['body'], str):
            headers = {'content-type': 'application/json'}
        else:
            headers = {}

        return cls(delay=delay, headers=headers, **data)


class Route(IdItem):
    """Route is a pair of request arguments and all possible reponses."""

    def __init__(
        self,
        id: str,
        responses: Iterable[Response],
        response_selection: ResponseSelectionStrategy,
        path: re.Pattern,
        auth: Auth,
        method: str = 'GET'
    ):
        super().__init__(id)
        self.response_selection = response_selection
        self.method = method
        self.path = path
        self.auth = auth
        self.used_count = 0
        self.responses: IdList[Response] = IdList()

        try:
            for response in responses:
                self.responses.add(response)
        except KeyError:
            raise DuplicateRouteError(f'Duplicate response id {response.id}.')

    def serialize(self) -> Dict[str, Any]:
        """Convert Route to JSON."""
        return {
            **super().serialize(),
            'response_selection': self.response_selection.serialize(),
            'auth': self.auth.serialize(),
            'method': self.method,
            'path': self.path.pattern,
            'used_count': self.used_count,
            'responses': self.responses.serialize(),
            'is_active': self.is_active
        }

    def get_response(self, response_id: str) -> Optional[Response]:
        """Get a Response by its id."""
        return self.responses.get(response_id)

    @classmethod
    def _create_responses(cls, responses: List[Dict[str, Any]]) -> List[Response]:
        """Create Response objects from json."""
        result = []
        for response in responses:
            if 'id' not in response:
                response['id'] = str(uuid.uuid4())
            result.append(Response.deserialize(response))
        return result

    @classmethod
    def deserialize(cls, data: Dict[str, Any]) -> Route:
        """Convert json to Route."""
        id = data.pop('id')
        auth = Auth.deserialize(data.pop('auth', None))
        path = re.compile(data.pop('path', None))
        response_selection = ResponseSelectionStrategy.deserialize(data.pop('response_selection', None))
        responses = cls._create_responses(data.pop('responses'))

        return cls(
            id=id,
            responses=responses,
            response_selection=response_selection,
            path=path,
            auth=auth,
            **data
        )

    def use(self, response: Response = None) -> None:
        """Increment use counter of this Route and given Response."""
        self.used_count += 1
        if response:
            response.use()

    def match(self, request: IncomingRequest) -> bool:
        """Return True, if this request specification matches given request and Route is active."""
        return all([
            self._match_method(request.method),
            self._match_path(request.path),
            self.is_active
        ])

    def _match_method(self, method: Optional[str]) -> bool:
        """Return True, if this requests HTTP method matches given InputRequest."""
        return self.method in [None, method]

    def _match_path(self, path: str) -> bool:
        """Return True, if this requests path matches given InputRequest."""
        return bool(self.path.match(path))

    def select_response(self) -> Optional[Response]:
        """Select response from list of responses."""
        return self.response_selection.select_response(self.responses)

    def authenticate(self, request: IncomingRequest) -> None:
        """Check if Request if properly authenticated."""
        self.auth.authenticate(request)

    @property
    def is_active(self) -> bool:
        """Return True if Route has at least one active Response."""
        for response in self.responses:
            if response.is_active:
                return True
        return False


class Router:
    """Custom request/response router."""

    def __init__(self) -> None:
        self.reset()

    def reset(self) -> None:
        """Remove all custom routes."""
        self.routes: IdList[Route] = IdList()

    def _generate_route_id(self) -> str:
        """Generate route id."""
        while (route_id := str(uuid.uuid4())) in self.routes:
            # It's virtually impossibe to generate the same uuid twice
            continue  # pragma: no cover
        return route_id

    def _set_route_id(self, route: Dict[str, Any], route_id: str = None) -> None:
        """Set route id if it doesn't already exist. Generate id if not set."""
        route.setdefault('id', route_id or self._generate_route_id())

    def add_route(self, route: Dict[str, Any]) -> Route:
        """Add custom request and matching responses."""
        self._set_route_id(route)
        route_object = Route.deserialize(route)
        try:
            self.routes.add(route_object)
        except KeyError:
            raise DuplicateRouteError(f'Route id "{route_object.id}" already exists.')
        return route_object

    def get_route(self, route_id: str) -> Optional[Route]:
        """Get Route by its id."""
        return self.routes.get(route_id)

    def remove_route(self, route_id: str) -> None:
        """Remove Route by its id."""
        self.routes.remove(route_id)

    def update_route(self, route: Dict[str, Any], route_id: str) -> Route:
        """Update route with completely new data."""
        self._set_route_id(route, route_id)
        if route_id != route['id'] and route['id'] in self.routes:
            raise DuplicateRouteError(
                f'Cannot change route id "{route_id}" to "{route["id"]}". Route id "{route["id"]}" already exists.'
            )

        route_object = Route.deserialize(route)
        try:
            self.routes.replace(route_id, route_object)
        except KeyError:
            raise MissingRouteError(f'Cannot update route "{route_id}". Route doesn\'t exist.')
        return route_object

    def match(self, incoming_request: IncomingRequest) -> Optional[Route]:
        """Find matching Route and return apropriet Response or None."""
        for route in self.routes:
            if route.match(incoming_request):
                return route
        return None



"""Usefull collections."""

import abc
from typing import Any, Dict, Generic, Iterator, List, Optional, TypeVar


class IdItem(abc.ABC):
    """Item containing id and serializable to json."""

    def __init__(self, id: str):
        self.id = id

    def serialize(self) -> Dict[str, Any]:
        """Serialize item to json."""
        return {'id': self.id}


IdItemType = TypeVar('IdItemType', bound=IdItem)


class IdList(Generic[IdItemType]):
    """List of Routes or Responses."""

    def __init__(self) -> None:
        self.items: List[IdItemType] = []

    def get(self, key: str) -> Optional[IdItemType]:
        """Get item with given key or None."""
        for item in self.items:
            if item.id == key:
                return item
        return None

    def __iter__(self) -> Iterator[IdItemType]:
        """Iterate over all items in list."""
        yield from self.items

    def serialize(self) -> List[Dict[str, Any]]:
        """Convert list to json."""
        return [item.serialize() for item in self]

    def __contains__(self, key: str) -> bool:
        """Return True if item with given key is present in list."""
        for item in self:
            if item.id == key:
                return True
        return False

    def __len__(self) -> int:
        """Get number of elements in list."""
        return len(self.items)

    def index(self, item_id: str) -> Optional[int]:
        """Get index of item in list of None."""
        for index, present_item in enumerate(self.items):
            if present_item.id == item_id:
                return index
        return None

    def add(self, item: IdItemType) -> None:
        """Add item to list."""
        if item.id in self:
            raise KeyError(f'Cannot insert item "{item.id}", key already exists.')
        self.items.append(item)

    def remove(self, item_id: str) -> None:
        """Remove item from list."""
        if (index := self.index(item_id)) is not None:
            del self.items[index]

    def replace(self, item_id: str, item: IdItemType) -> None:
        """Replace item on given index with new item."""
        if (index := self.index(item_id)) is None:
            raise KeyError(f'Item id "{item_id}" does\'t exist.')
        self.items[index] = item



"""Install Trickster and dependencies."""

import os

from setuptools import setup


def get_version() -> str:
    """Get current package version."""
    return os.environ.get('PACKAGE_VERSION', 'dev')


def get_long_description() -> str:
    """Get long description of package."""
    pwd = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(pwd, 'README.md'), encoding='utf-8') as f:
        return f.read()


def main() -> None:
    """Build Trickster package."""
    setup(
        name='trickster',
        version=get_version(),
        python_requires='>=3.8',
        description='Trickester is a Python/Flask application for mocking REST APIs',
        long_description=get_long_description(),
        long_description_content_type='text/markdown',
        url='https://github.com/JakubTesarek/trickster',
        author='Jakub Tesarek',
        author_email='jakub@tesarek.me',
        license='MIT',
        include_package_data=True,
        packages=['trickster'],
        install_requires=[
            'click',
            'gunicorn',
            'flask',
            'fastjsonschema',
            'basicauth'
        ],
        extras_require={
            'dev': [
                'setuptools',
                'wheel',
                'flake8',
                'flake8-docstrings',
                'flake8-pytest',
                'flake8-eradicate',
                'flake8-print',
                'flake8-todo',
                'flake8-import-order',
                'pytest',
                'pytest-repeat',
                'pytest-cov',
                'pytest-mock',
                'mypy',
                'twine'
            ]
        },
        entry_points={
            'console_scripts': [
                'trickster=cli:cli'
            ]
        }
    )


if __name__ == '__main__':
    main()



"""Initialization of API app."""

from __future__ import annotations

from typing import Any, Optional, Tuple

from flask import Flask, jsonify

from trickster.endpoints import external_api, internal_api
from trickster.routing import Router

from werkzeug.exceptions import HTTPException


PORT = 5000
INTERNAL_PREFIX = '/internal'


def http_error_handler(error: HTTPException) -> Tuple[Any, Optional[int]]:
    """Handler for error pages."""
    return jsonify({
        'error': error.name,
        'message': error.description
    }), error.code


class ApiApp(Flask):
    """Flask application handling API endpoints."""

    def __init__(self, internal_prefix: str = INTERNAL_PREFIX) -> None:
        super().__init__(__name__)
        self.internal_prefix = internal_prefix
        self.user_router = Router()
        self._register_handlers()
        self._register_blueprints()

    def _register_handlers(self) -> None:
        """Register error page handlers."""
        self.register_error_handler(HTTPException, http_error_handler)

    def _register_blueprints(self) -> None:
        """Register api endpoints."""
        self.register_blueprint(internal_api, url_prefix=self.internal_prefix)
        self.register_blueprint(external_api, url_prefix='')



"""API endpoints."""

from __future__ import annotations

from flask import Blueprint, Response, abort, current_app, jsonify, make_response, request

from trickster import RouteConfigurationError
from trickster.auth import AuthenticationError
from trickster.input import HTTP_METHODS, IncomingFlaskRequest, IncomingTestRequest
from trickster.validation import request_schema


internal_api = Blueprint('internal_api', __name__)
external_api = Blueprint('external_api', __name__)


@internal_api.route('/health', methods=['GET'])
def health() -> Response:
    """Returns internal app status."""
    return jsonify({'status': 'ok'})


@internal_api.route('/routes', methods=['GET'])
def get_all_routes() -> Response:
    """Get list of configured Routes."""
    return jsonify(current_app.user_router.routes.serialize())


@internal_api.route('/routes', methods=['POST'])
@request_schema('route.schema.json')
def add_route() -> Response:
    """Create new route."""
    try:
        route = current_app.user_router.add_route(request.get_json())
        return make_response(jsonify(route.serialize()), 201)
    except RouteConfigurationError as error:
        abort(error.http_code, str(error))


@internal_api.route('/routes', methods=['DELETE'])
def remove_all_routes() -> Response:
    """Reset router configuration."""
    current_app.user_router.reset()
    return make_response('', 204)


@internal_api.route('/routes/<string:route_id>', methods=['GET'])
def get_route(route_id: str) -> Response:
    """Get route by id."""
    if route := current_app.user_router.get_route(route_id):
        return make_response(jsonify(route.serialize()), 200)
    abort(404, f'Route id "{route_id}" does not exist.')


@internal_api.route('/routes/<string:route_id>', methods=['PUT'])
@request_schema('route.schema.json')
def replace_route(route_id: str) -> Response:
    """Replace route with new data."""
    try:
        route = current_app.user_router.update_route(request.get_json(), route_id)
        return make_response(jsonify(route.serialize()), 201)
    except RouteConfigurationError as error:
        abort(error.http_code, str(error))


@internal_api.route('/routes/<string:route_id>', methods=['DELETE'])
def remove_route(route_id: str) -> Response:
    """Remove route by id."""
    if current_app.user_router.get_route(route_id):
        current_app.user_router.remove_route(route_id)
        return make_response('', 204)
    abort(404, f'Route id "{route_id}" does not exist.')


@internal_api.route('/match_route', methods=['POST'])
@request_schema('request.schema.json')
def match_route() -> Response:
    """Match configured routes against given request."""
    payload = request.get_json()
    incoming_request = IncomingTestRequest(
        base_url=request.host_url,
        full_path=payload['path'],
        method=payload['method']
    )

    if route := current_app.user_router.match(incoming_request):
        return make_response(jsonify(route.serialize()), 200)
    abort(404, 'No route was matched.')


@internal_api.route('/routes/<string:route_id>/responses', methods=['GET'])
def get_all_responses(route_id: str) -> Response:
    """Get all responses from given route."""
    if route := current_app.user_router.get_route(route_id):
        return make_response(jsonify(route.responses.serialize()), 200)
    abort(404, f'Route id "{route_id}" does not exist.')


@internal_api.route('/routes/<string:route_id>/responses/<string:response_id>', methods=['GET'])
def get_response(route_id: str, response_id: str) -> Response:
    """Get response by id from given route.."""
    if route := current_app.user_router.get_route(route_id):
        if response := route.get_response(response_id):
            return make_response(jsonify(response.serialize()), 200)
        abort(404, f'Response id "{response_id}" does not exist in request id "{route_id}".')
    abort(404, f'Route id "{route_id}" does not exist.')


@external_api.route('/<path:path>', methods=HTTP_METHODS)
def respond(path: str) -> Response:
    """Match request againts defined routes and return appropriet response."""
    incomming_request = IncomingFlaskRequest(request)
    try:
        if route := current_app.user_router.match(incomming_request):
            route.authenticate(incomming_request)
            response = route.select_response()
            response.wait()
            route.use(response)
            return response.as_flask_response()
        abort(404)
    except AuthenticationError as error:
        abort(401, str(error))
